#include <Windows.h>
#include <Stdio.h>
#include <math.h>
#include <d3d8types.h>
#include <d3dx8.h>
#include "Main.h"
#include "KUtil.h"
#include "Logg.h"
#include "StringTable.h"
#include "MemLeak.h"
#include "FastMath.h"
#include "Vectors.h"
#include "Directory.h"
#include "GameFont.h"
#include "Lights.h"
#include "PlaneBlendVS.h"
#include "PlaneBlend.h"

extern EngineData	g_ed ;
#include "CommonDecls.h"


/////////////////////  Local prototypes /////////////////////
static void __fastcall LoadPlaneBlendConstants(IDirect3DDevice8* pdevice, D3DXMATRIX* world, 
								D3DXMATRIX* LeftPlaneMat, D3DXMATRIX* RightPlaneMat) ;
static BOOL CreateShadersForPlaneBlend(IDirect3DDevice8* pdevice) ;
///////////////////////////////////////////////////////////////

///////////////////////// Globals /////////////////////////////

LPDIRECT3DVERTEXBUFFER8		g_pbVB ;	// vertex buffer which holds our plane vertex data
LPDIRECT3DINDEXBUFFER8		g_pbIB ;	// index buffer which holds our plane face data
LPDIRECT3DTEXTURE8			g_pbTexture ;// texture

#define MAX_V_SHADERS			5

#define BASIC_VS				0		// VS for transforming vertex
#define BASIC_VS_WITHLIGHT		1		// VS for transforming and lighting
#define BLEND_ONLY				2		// VS for blending the vertex.
#define NON_BLEND_ONLY			3		// VS for no-blending
#define BLEND_WITH_LIGHT		4		// VS for blending with lighting.

// shader handles
DWORD						g_pbVS[MAX_V_SHADERS] ;	// array our shader handles.


DWORD						g_dwpbNumVertices ,	// Number of vertices in the plane
							g_dwpbNumIndices ;	// Number of faces in the plane

D3DMATRIX					g_rotMatPlane1 ,	// rotation matrix for the left half of the plane
							g_rotMatPlane2 ;	// rotation matrix for the right half of the plane
///////////////////////////////////////////////////////////////



#define MAX_PLANE_TESS_LEVEL_HORZ	5	// horizontal tesselation level
#define MAX_PLANE_TESS_LEVEL_VERT	3	// vertical tesselation level
#define PLANE_WIDTH		40				// physical width of the plane
#define PLANE_HEIGHT	30				// physical height of the plane



BOOL InitPlaneBlend(IDirect3DDevice8* pdevice)
{
//////////////////////// Local variables ////////////////////////
	BOOL		res = TRUE ;
	VertexB2	locPlaneData[((MAX_PLANE_TESS_LEVEL_HORZ*2)+1) * ((MAX_PLANE_TESS_LEVEL_VERT*2)+1)] ;
	Face		locFaceData[((MAX_PLANE_TESS_LEVEL_HORZ*2)*(MAX_PLANE_TESS_LEVEL_VERT*2))*4] ;
	DWORD		i, j ;
	float		fHorzPosShift, fVertPosShift ;
	float		fHorzTextureCoordShift ;
	float		fVertTextureCoordShift ;
	float		tu, tv ;
	DWORD		iIndex ;
	D3DVECTOR	Position = { 0.0f } ;
	D3DVECTOR	Normal = { 0.0f, 0.0f, -1.0f } ;
	DWORD		dwNumIndices, dwNumVertices ;
	DWORD		a,b,c, k ;
	DWORD		dwCount1, dwCount2 ;

	float		fWeightStartHorz, fWeightStartVert, fWeightShiftHorz, fWeightShiftVert ;
////////////////////////////////////////////////////////////


	LogMsg("\n In InitPlaneBlend \n") ;


	if(!SwitchToWorldFolder())
	{
		e_xit ;
	}

/*	The following code sets up the plane data. This has been explained --
	-- in detail on the web page.
	It's just plain linear math. Shouldn't be too much of a problem.
	Hope my code is correct.

	What I am doing here is linearly interpolate Position, texture co-ordinates,
	and blending weights based on the tesselation level and physical parameters of the plane.
*/

	ZeroMemory(locPlaneData, sizeof(locPlaneData)) ;
	ZeroMemory(locFaceData, sizeof(locFaceData)) ;

	fHorzTextureCoordShift = 0.5f / ((((float)MAX_PLANE_TESS_LEVEL_HORZ)+0.0f)) ;
	fVertTextureCoordShift = 0.5f / ((((float)MAX_PLANE_TESS_LEVEL_VERT)+0.0f)) ;

	fHorzPosShift = (((FLOAT)PLANE_WIDTH) * 0.5f) / ((float)MAX_PLANE_TESS_LEVEL_HORZ) ;
	fVertPosShift = (((FLOAT)PLANE_HEIGHT) * 0.5f) / ((float)MAX_PLANE_TESS_LEVEL_VERT) ;

	fWeightShiftHorz = (0.5f) / ((float)MAX_PLANE_TESS_LEVEL_HORZ) ;
	fWeightShiftVert = (0.0f) ;

	tu = 0.0f ;
	tv = 0.0f ;
	iIndex = 0 ;
	Position.x = -((((FLOAT)PLANE_WIDTH) * 0.5f)) ;
	Position.y = ((((FLOAT)PLANE_HEIGHT) * 0.5f)) ;
	Position.z = 0.0f ;

	g_dwpbNumVertices = dwNumVertices = ((MAX_PLANE_TESS_LEVEL_HORZ*2)+1) * ((MAX_PLANE_TESS_LEVEL_VERT*2)+1) ;
	dwCount1 = ((MAX_PLANE_TESS_LEVEL_VERT*2)+1) ;
	dwCount2 = ((MAX_PLANE_TESS_LEVEL_HORZ*2)+1) ;
	for(i=0;i<dwCount1;i++)
	{
		fWeightStartHorz = 1.0f ;
		fWeightStartVert = 0.0f ;
		for(j=0;j<dwCount2;j++,iIndex++)
		{
			locPlaneData[iIndex].Position = Position ;
			locPlaneData[iIndex].Normal = Normal ;

			locPlaneData[iIndex].tu = tu + (((float)j)*fHorzTextureCoordShift) ;
			locPlaneData[iIndex].tv = tv + (((float)i)*fVertTextureCoordShift) ;

			if(fWeightStartHorz >= 0.5f)
			{
				locPlaneData[iIndex].Weights[0] = fWeightStartHorz ;
				locPlaneData[iIndex].Weights[1] = 1.0f-fWeightStartHorz ;

				locPlaneData[iIndex].Indices[0] = (BYTE)(0) ;
				locPlaneData[iIndex].Indices[1] = (BYTE)(1) ;
			}
			else
			{
				locPlaneData[iIndex].Weights[1] = fWeightStartHorz ;
				locPlaneData[iIndex].Weights[0] = 1.0f-fWeightStartHorz ;

				locPlaneData[iIndex].Indices[1] = (BYTE)(0) ;
				locPlaneData[iIndex].Indices[0] = (BYTE)(1) ;
			}

			fWeightStartHorz -= fWeightShiftHorz ;

			Position.x += (fHorzPosShift) ;
		}
		Position.x = -((((FLOAT)PLANE_WIDTH) * 0.5f)) ;
		Position.y += -(fVertPosShift) ;
	}

	g_dwpbNumIndices = dwNumIndices = ((MAX_PLANE_TESS_LEVEL_HORZ*2)*(MAX_PLANE_TESS_LEVEL_VERT*2))*2 ;
	dwCount1 = (MAX_PLANE_TESS_LEVEL_VERT*2) ;
	dwCount2 = (MAX_PLANE_TESS_LEVEL_HORZ*2)*2 ;
	for(i=0;i<dwCount1;i++)
	{
		k = 0 ;
		for(j=0;j<dwCount2;j+=2,k++)
		{
			a = (k) + ((i) * ((MAX_PLANE_TESS_LEVEL_HORZ*2)+1)) ;
			b = (k+1) + ((i) * ((MAX_PLANE_TESS_LEVEL_HORZ*2)+1)) ;
			c = (k) + ((i+1) * ((MAX_PLANE_TESS_LEVEL_HORZ*2)+1)) ;

			iIndex = (i*dwCount2) + j ;

			locFaceData[iIndex].v[0] = (WORD)a ;
			locFaceData[iIndex].v[1] = (WORD)b ;
			locFaceData[iIndex].v[2] = (WORD)c ;

			locFaceData[iIndex+1].v[0] = (WORD)b ;
			locFaceData[iIndex+1].v[1] = (WORD)(c+1) ;
			locFaceData[iIndex+1].v[2] = (WORD)(c) ;
		}
	}

	/////////// Print plane Debug info ///////////////
	LogMsg("\n Plane Info: \n") ;
	LogMsg("\n ------- Vertex Info -------\n") ;
	LogMsg("\n Num Vertices = %d\n", g_dwpbNumVertices) ;
	for(i=0;i<g_dwpbNumVertices;i++)
	{
		LogMsg("\n Vertex [%03d] : %-5.3f, %-5.3f, %-5.3f", i, locPlaneData[i].Position.x, locPlaneData[i].Position.y, locPlaneData[i].Position.z) ;
		LogMsg("  \t%-4.3f, %-4.3f, Indices = %02d, %02d,   Tex = %-4.3f, %-4.3f", locPlaneData[i].Weights[0], locPlaneData[i].Weights[1], 
			locPlaneData[i].Indices[0], locPlaneData[i].Indices[1], 
			locPlaneData[i].tu, locPlaneData[i].tv) ;

		if(i && (i%((MAX_PLANE_TESS_LEVEL_HORZ*2)))==0)
		{
			LogMsg("\n--- row ---\n") ;
		}
	}

	LogMsg("\n--- End of vertex info ---\n") ;

	LogMsg("\n ----- Face Info ----- \n") ;
	LogMsg("\n Num Faces = %d\n", g_dwpbNumIndices) ;
	for(i=0;i<g_dwpbNumIndices;i++)
	{
		LogMsg("\n Face [%03d] : %d, %d, %d", i, locFaceData[i].v[0], locFaceData[i].v[1], locFaceData[i].v[2]) ;
	}
	/////////////////////////////////////////////


	// create the vertex buffer which holds the plane data
	if(!InitVertexBuffer(&g_pbVB, pdevice, locPlaneData, sizeof(locPlaneData),
		D3DFVF_PB_VERTEXFORMAT, 0, D3DPOOL_DEFAULT))
	{
		LogMsg("\n Failed to Create Vertex Buffer\n") ;
		e_xit ;
	}

	// create the index buffer which holds the plane data
	if(!InitIndexBuffer(&g_pbIB, pdevice, locFaceData, sizeof(locFaceData),
		D3DFMT_INDEX16, 0, D3DPOOL_DEFAULT))
	{
		LogMsg("\n Failed to Create Index Buffer\n") ;
		e_xit ;
	}

	// load our texture
	if(!LoadTexture(&g_pbTexture, pdevice, _T("Plane.tga"), FALSE, TF_32BITRGB, TRUE))
	{
		LogMsg("\n Failed to Create texture \n") ;
		e_xit ;
	}

	// this is where all the different shaders are loaded and created.
	if(!CreateShadersForPlaneBlend(pdevice))
	{
		LogMsg("\n CreateShadersForPlaneBlend FAILED\n") ;
		e_xit ;
	}

	// make the matrices identity.
	D3DXMatrixIdentity(&g_rotMatPlane1) ;
	D3DXMatrixIdentity(&g_rotMatPlane2) ;

	// set default shader
	SETVERTEXSHADER(pdevice, g_pbVS[BASIC_VS]) ;

	LogMsg("\n Out of InitPlaneBlend successfully \n") ;

xit :
	return res ;
}


/*
	g_ed.lastKey is the last windows key pressed.

	You may have to change this variable(s) to suit your app.
*/
static void UpdatePlaneRotationMatrices(D3DMATRIX* mat1, D3DMATRIX* mat2)
{
	static float	fRotPlane1 ;
	static float	fRotPlane2 ;
	const float		fSkip = 2.0f ;

	switch(g_ed.lastKey)
	{
		case VK_F1 :
			fRotPlane1 += fSkip ;
		break ;

		case VK_F2 :
			fRotPlane1 -= fSkip ;
		break ;

		case VK_F3 :
			fRotPlane2 += fSkip ;
		break ;
	
		case VK_F4 :
			fRotPlane2 -= fSkip ;
		break ;
	}

	// rotate on Y Axis only
	SetYRotate(mat1, D3DXToRadian(fRotPlane1)) ;
	SetYRotate(mat2, D3DXToRadian(fRotPlane2)) ;
}


/* This procedure sets up a new vertexshader based on certain keys being pressed
*/
static DWORD SetPlaneBlendVS(IDirect3DDevice8* device8)
{
	static DWORD	dwVSIndex = NON_BLEND_ONLY ;
	
	switch(g_ed.lastKey)
	{
		case '1' :
			dwVSIndex = BASIC_VS ;
		break ;

		case '2' :
			dwVSIndex = BASIC_VS_WITHLIGHT ;
		break ;

		case '3' :
			dwVSIndex = BLEND_ONLY ;
		break ;

		case '4' :
			dwVSIndex = NON_BLEND_ONLY ;
		break ;

		case '5' :
			dwVSIndex = BLEND_WITH_LIGHT ;
		break ;
	}
	
	SETVERTEXSHADER(device8, g_pbVS[dwVSIndex]) ;

	return dwVSIndex ;
}


static void DisplayPlaneBlendDebugText(const DWORD dwVSIndex)
{
	static LPCTSTR	lpszString = _T("Non-Blend only") ;
	static LPCTSTR	lpszStringOne = _T("Basic transformation only") ;
	static LPCTSTR	lpszStringTwo =  _T("Basic vertex lit surface only") ;
	static LPCTSTR	lpszStringThree = _T("Plane transformation with blend WITHOUT light") ;
	static LPCTSTR	lpszStringFour = _T("Plane transformation WITHOUT blend") ;
	static LPCTSTR	lpszStringFive = _T("Plane transformation with blend with light") ;

	switch(dwVSIndex)
	{
		case BASIC_VS :
			lpszString = lpszStringOne ;
		break ;

		case BASIC_VS_WITHLIGHT :
			lpszString = lpszStringTwo ;
		break ;

		case BLEND_ONLY :
			lpszString = lpszStringThree ;
		break ;

		case NON_BLEND_ONLY :
			lpszString = lpszStringFour ;
		break ;

		case BLEND_WITH_LIGHT :
			lpszString = lpszStringFive ;
		break ;
	}
	
	DrawImmediateText(1, 50, 0, DOARGB(255,255,255,255), 0, "Vertex shader key Map") ;
	DrawImmediateText(1, 75, 0, DOARGB(255,255,255,255), 0, "1 = %s", lpszStringOne) ;
	DrawImmediateText(1, 100, 0, DOARGB(255,255,255,255), 0, "2 = %s", lpszStringTwo) ;
	DrawImmediateText(1, 125, 0, DOARGB(255,255,255,255), 0, "3 = %s", lpszStringThree) ;
	DrawImmediateText(1, 150, 0, DOARGB(255,255,255,255), 0, "4 = %s", lpszStringFour) ;
	DrawImmediateText(1, 180, 0, DOARGB(255,255,255,255), 0, "5 = %s", lpszStringFive) ;


	DrawImmediateText(1, 30, 0, DOARGB(255,255,0,255), 0, "Current shader = %s", lpszString) ;
}


/*
	PROCEDURE DisplayPlaneBlend.
	
	void DisplayPlaneBlend(IDirect3DDevice8* device8) ;
	This procedure is called every loop. It does the following things:
	1. Updates the plane rotation matrices based on keyboard input.
	2. Sets the world matrix to identity.
	3. Sets the vertex shader based on keyboard input.
	4. Loads the verex shader constants. These constants are independant of
		the different vertex shaders.
	5. Binds the stream source to the plane vertex buffer and the 
		indices to the plane index buffer. Sets the current texture for the stage 
		to the plane texture.
	6. Finally calls DrawIndexedPrimitive.

	You will find the macros for DRAWIP, SETINDICES and others in the file
	UtilDefines.h
*/
void DisplayPlaneBlend(IDirect3DDevice8* device8)
{
	D3DMATRIX	idmat ;
	DWORD		dwIndex ;

	// update rotation matrices based on certain key press events.
	UpdatePlaneRotationMatrices(&g_rotMatPlane1, &g_rotMatPlane2) ;

	// identity world matrix
	D3DXMatrixIdentity((D3DXMATRIX*)&idmat) ;
	MAKETRANSMAT(idmat, 0.0f, 35.0f, 0.0f) ;

	dwIndex = SetPlaneBlendVS(device8) ;
	LoadPlaneBlendConstants(device8, &idmat, &g_rotMatPlane1, &g_rotMatPlane2) ;

	// do not disable lighting on the plane.
	SETTEXTURESTAGE(device8, 0, D3DTSS_COLOROP,   D3DTOP_MODULATE) ;
	SETTEXTURESTAGE(device8, 0, D3DTSS_COLORARG1, D3DTA_TEXTURE) ;
	SETTEXTURESTAGE(device8, 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE) ;
	
	SETSTREAMSOURCE(device8, 0, g_pbVB, SIZEOFVERTEXB2) ;
	SETINDICES(device8, g_pbIB, 0) ;
	SETTEXTURE(device8, 0, NULL) ;
	SETTEXTURE(device8, 0, g_pbTexture) ;

	DRAWIP(device8, D3DPT_TRIANGLELIST, 0, g_dwpbNumVertices, 0, g_dwpbNumIndices) ;

	DisplayPlaneBlendDebugText(dwIndex) ;
}



/*
	This procedure creates shaders for the different effects.
	Note that the declaration remains the same for all of them.
*/
static BOOL CreateShadersForPlaneBlend(IDirect3DDevice8* pdevice)
{
	BOOL		res = TRUE ;
	LPCTSTR		lpszShaders[] = {	_T("BasicTr.vso") , 
									_T("BasicTrWithLight.vso") ,
									_T("BlendVS.vso") ,
									_T("BlendNoVS.vso") ,
									_T("BlendWithLight.vso") , } ;
	LPCTSTR		lpszStr ;
	DWORD		declBasicTr[] = 
						{
						D3DVSD_STREAM(0),
						D3DVSD_REG(D3DVSDE_POSITION, D3DVSDT_FLOAT3 ),		// position
						D3DVSD_REG(D3DVSDE_BLENDWEIGHT, D3DVSDT_FLOAT2 ),	// 2 blending weights
						D3DVSD_REG(D3DVSDE_BLENDINDICES, D3DVSDT_UBYTE4 ),	// 4 indices, (we need only 2). you cn use this or use D3DVSDT_SHORT2 and change the data type in the VertexB2 structure
						D3DVSD_REG(D3DVSDE_NORMAL, D3DVSDT_FLOAT3 ),		// normal for light calcula
						D3DVSD_REG(D3DVSDE_TEXCOORD0,D3DVSDT_FLOAT2 ),		// texture co-ordinates
						D3DVSD_END()
					};
	DWORD		i ;

	for(i=0;i<sizeof(lpszShaders)/sizeof(lpszShaders[0]);i++)
	{
		lpszStr = lpszShaders[i] ;
		if(!LoadAndCreateShader(pdevice, lpszStr, declBasicTr, 0, SHADERTYPE_VERTEX,
			&g_pbVS[i]))
		{
			LogMsg("\n\n LoadAndCreateShader with [%s] FAILED \n", lpszStr) ;
			e_xit ;
		}
	}

xit :
	return res ;
}

/*
	Procedure:	void LoadPlaneBlendConstants(IDirect3DDevice8* pdevice, D3DXMATRIX* world, 
				D3DXMATRIX* LeftPlaneMat, D3DXMATRIX* RightPlaneMat) ;
	Return:		-- None --

	Desc:		Loads the constants required by the plane vertex shader to VS constant registers.
				Constants include light data.
*/

/*
	You may have to meddle around with this procedure. I am using my own 
	data structures for light data. Maybe you can replace them with your own.
	
	All I need for the light data is it's world position, color, intensity, and attenuation.
	also final color = (color*intensity).
	
*/
static void __fastcall LoadPlaneBlendConstants(IDirect3DDevice8* pdevice, D3DXMATRIX* world, 
								D3DXMATRIX* LeftPlaneMat, D3DXMATRIX* RightPlaneMat)
{
	D3DXMATRIX		trans, matTransposed, TransposedWorld ;
	D3DXMATRIX		LeftPlaneMatTransposed, RightPlaneMatTransposed ;
	D3DXMATRIX		view, proj ;
	float			fUpload[4*12] = { 1.0f } ;
	float			fUploadFourZeroes[4] = { 0.0f } ;
	D3DLIGHT8*		light8 ;
	extern LightHdr	g_lightHdr ;
	LightProp*		ltProp ;
	float			c4[4] = { 4.0f, CV_BONE_MATRICES_START_FLT, 1.0f, 0.0f } ;

	pdevice->lpVtbl->GetTransform(pdevice, D3DTS_VIEW, &view) ;
	pdevice->lpVtbl->GetTransform(pdevice, D3DTS_PROJECTION, &proj) ;

	// view x projection
	D3DXMatrixMultiply(&trans, &view, &proj) ;
	
	D3DXMatrixTranspose(&matTransposed, &trans) ;
	D3DXMatrixTranspose(&TransposedWorld, (D3DXMATRIX*)world) ;

	// c0 = utility constants
	pdevice->lpVtbl->SetVertexShaderConstant(pdevice, 0, c4, 1) ;

	// c1 = Zeroes
	pdevice->lpVtbl->SetVertexShaderConstant(pdevice, 1, fUploadFourZeroes, 1) ;

	// c14 to c17 = world matrix
	pdevice->lpVtbl->SetVertexShaderConstant(pdevice, CV_WORLD_0, &TransposedWorld, 4) ;

	// c18 to c21 = View x Projection
	pdevice->lpVtbl->SetVertexShaderConstant(pdevice, CV_VIEWPROJ_0, &matTransposed, 4) ;


	light8 = GetLightProperties(0) ;
	ltProp = &g_lightHdr.lp[0] ;
	
	// c2 to c4 = light 1
	CopyMemory(&fUpload[0], &light8->Position, sizeof(D3DVECTOR)) ;
	CopyMemory(&fUpload[4], &light8->Attenuation0, sizeof(D3DVECTOR)) ;
	fUpload[8] = light8->Diffuse.r * ltProp->rIntensity ;
	fUpload[9] = light8->Diffuse.g * ltProp->gIntensity ;
	fUpload[10] = light8->Diffuse.b * ltProp->bIntensity ;
	fUpload[11] = 1.0f ;

	light8 = GetLightProperties(1) ;
	ltProp = &g_lightHdr.lp[1] ;
	// c5 to c7 = light 2
	CopyMemory(&fUpload[12], &light8->Position, sizeof(D3DVECTOR)) ;
	CopyMemory(&fUpload[16], &light8->Attenuation0, sizeof(D3DVECTOR)) ;
	fUpload[20] = light8->Diffuse.r * ltProp->rIntensity ;
	fUpload[21] = light8->Diffuse.g * ltProp->gIntensity ;
	fUpload[22] = light8->Diffuse.b * ltProp->bIntensity ;
	fUpload[23] = 1.0f ;

	// upload light data
	pdevice->lpVtbl->SetVertexShaderConstant(pdevice, 2, fUpload, 12) ;

	
	D3DXMatrixTranspose(&LeftPlaneMatTransposed, (D3DXMATRIX*)LeftPlaneMat) ;
	D3DXMatrixTranspose(&RightPlaneMatTransposed, (D3DXMATRIX*)RightPlaneMat) ;

	// upload the left and right transposed plane rotation matrices
	pdevice->lpVtbl->SetVertexShaderConstant(pdevice, CV_PLANE_BLEND_LEFT_MATRIX, 
		&LeftPlaneMatTransposed, 4) ;
	pdevice->lpVtbl->SetVertexShaderConstant(pdevice, CV_PLANE_BLEND_RIGHT_MATRIX, 
		&RightPlaneMatTransposed, 4) ;
}



/*
	Cleans up the resources.
*/
void CleanupPlaneBlend(IDirect3DDevice8* pdevice)
{
	DWORD	i ;

	LogMsg("\n In CleanupPlaneBlend \n") ;

	if(pdevice)
	{
		for(i=0;i<MAX_V_SHADERS;i++)
		{
			if(g_pbVS)
			{
				pdevice->lpVtbl->DeleteVertexShader(pdevice, g_pbVS[i]) ;
				g_pbVS[i] = 0 ;
			}
		}
	}

	SAFE_RELEASE(g_pbVB) ;
	SAFE_RELEASE(g_pbIB) ;
	SAFE_RELEASE(g_pbTexture) ;

	// uncomment the following line.
//	CloseLogFile() ;
}