#include <Windows.h>
#include <stdio.h>
#include <TChar.h>
#include <d3d8.h>
#include <d3d8types.h>
#include <d3dx8.h>
#include <dxerr8.h>

#include "Utils.h"

void * __cdecl ALLOCMEM(unsigned int memsize)
{
	void *ptr = (void *)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, memsize) ;
	if(NULL==ptr)
	{
		LogMsg("\n Failed to allocate memory\n") ;
	}

	return(ptr);
}

#define FREEMEM(p) { if(p) HeapFree(GetProcessHeap(), 0, p) , p=0 ; }


// Loads a *.vso file and creates a vertex shader for it
BOOL LoadAndCreateShader(LPDIRECT3DDEVICE8 m_pD3DDev, LPCTSTR filename, const DWORD* pDeclaration, 
							DWORD Usage, DWORD ShaderType, DWORD* pHandle)
{
	BOOL			res = TRUE ;
	HANDLE			hFile = 0 ;
	HRESULT			hr ;
	DWORD			dwFileSize ;
	DWORD*			pShader = 0 ;
	char			szBuffer[128] = { 0 } ;

	if(!m_pD3DDev)
	{
		e_xit ;
	}

	hFile = CreateFile(filename, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
	if(hFile == INVALID_HANDLE_VALUE)
	{
		LogMsg("\n\nFailed to open file\n\n") ;
		e_xit ;
	}

	dwFileSize = GetFileSize(hFile, NULL);

	pShader = (DWORD*)ALLOCMEM(dwFileSize) ;
	if (!pShader)
	{
		e_xit ;
	}

	ReadFile(hFile, (void*)pShader, dwFileSize, &dwFileSize, NULL);

	if (ShaderType == SHADERTYPE_VERTEX)
	{
		hr = m_pD3DDev->lpVtbl->CreateVertexShader(m_pD3DDev, pDeclaration, pShader, pHandle, Usage) ;
		if ( FAILED(hr) )
		{
			LPTSTR	lpStr1 = _T("\n\nFailed to create VertexShader, errors:\n") ;
			OutputDebugString(lpStr1) ;
			LogMsg(lpStr1) ;
			D3DXGetErrorStringA(hr , szBuffer, sizeof(szBuffer));
			OutputDebugString(szBuffer);
			OutputDebugString("\n\n\n");
			LogMsg("\n%s\n\n\n", szBuffer) ;
			e_xit ;
		}
	}
	else
	{
		if (ShaderType == SHADERTYPE_PIXEL)
		{
			hr = m_pD3DDev->lpVtbl->CreatePixelShader(m_pD3DDev, pShader, pHandle) ;
			if ( FAILED(hr) )
			{
				LPTSTR	lpStr1 = _T("\n\nFailed to create PixelShader, errors:\n") ;
				OutputDebugString(lpStr1) ;
				LogMsg(lpStr1) ;
				D3DXGetErrorStringA(hr , szBuffer, sizeof(szBuffer));
				OutputDebugString(szBuffer);
				OutputDebugString("\n\n\n");
				LogMsg("\n%s\n\n\n", szBuffer) ;
				e_xit ;
			}
		}
		else
		{
			LogMsg("\n FAILED!!! ERROR!!!! Invalid shader type \n\n") ;
			e_xit ;
		}
	}

xit :

	if(pShader)
	{
		FREEMEM((void*)pShader) ;
		pShader = 0 ;
	}


	if(hFile)
	{
		CloseHandle(hFile) ;
		hFile = 0 ;
	}

	return res ;
}


BOOL LoadTexture(LPDIRECT3DTEXTURE8* ppTexture, LPDIRECT3DDEVICE8 device8, 
				 LPCTSTR lpszFilename, const BOOL bUseMipLevels,
				 const DWORD dwTextureFormat, const BOOL bCompressed)
{
	BOOL				res = TRUE ;
	LPDIRECT3DTEXTURE8	lpTex = 0 ;
	D3DFORMAT           TextureFormat ;
	HRESULT				hr ;
	TCHAR				szBuffer[256] ;
	LPDIRECT3DSURFACE8	pBackBuffer = 0 ;
	D3DSURFACE_DESC		bbSurfaceDesc ;

	switch(dwTextureFormat)
	{
		case TF_16BITRGB :
			TextureFormat = D3DFMT_R5G6B5 ;
		break ;

		case TF_32BITRGB_ALPHA :
			TextureFormat = D3DFMT_A8R8G8B8 ;
		break ;

		case TF_32BITRGB :
		case TF_24BITRGB :
			TextureFormat = D3DFMT_X8R8G8B8 ;
		break ;

		default :
			// Store render target surface desc
			hr = device8->lpVtbl->GetBackBuffer(device8, 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer) ;
			if(FAILED(hr))
			{
				GetDXErrorEx(" GetBackBuffer for device8 FAILED. ", hr) ;
				e_xit ;
			}

			hr = pBackBuffer->lpVtbl->GetDesc(pBackBuffer, &bbSurfaceDesc) ;
			if(FAILED(hr))
			{
				GetDXErrorEx(" GetDesc for BackBuffer for device8 FAILED. ", hr) ;
				e_xit ;
			}
			TextureFormat = bbSurfaceDesc.Format ;
		break ;
	}
	
	hr = D3DXCreateTextureFromFileExA(device8, lpszFilename, 0, 0, 1, 0, TextureFormat, 
		D3DPOOL_DEFAULT, D3DX_FILTER_TRIANGLE | D3DX_FILTER_DITHER, 
		D3DX_FILTER_BOX , 0, NULL, NULL, &lpTex) ;
	if(FAILED(hr))
	{
		LogMsg(" D3DXCreateTextureFromFileExA for file <%s> FAILED. ", lpszFilename) ;
		GetDXErrorEx(szBuffer, hr) ;
		LogMsg("  Error = %s\n\\n", szBuffer) ;
		e_xit ;	
	}

xit :
	if(!res)
	{
		SAFE_RELEASE(lpTex) ;
	}
	SAFE_RELEASE(pBackBuffer) ;
	*ppTexture = lpTex ;
	return res ;
}


const TCHAR* GetDXErrorEx(LPCTSTR start, const HRESULT hr)
{
	TCHAR	szBuf[1024] ;
	LPCSTR	lpstr = DXGetErrorString8(hr) ;

	szBuf[0] = 0 ;
	wsprintf(szBuf, "\n %s . [[ %s ]]\n", start, lpstr) ;
	LogMsg(szBuf) ;

	return lpstr ;
}


static FILE*	g_LogFilePtr ;
static LPCTSTR	g_lpszLogFileName = _T("NewLogg.txt") ;
void LogMsg(char *fmt,...)
{
	char buff[1024] ;

	va_list argPtr ;
	va_start(argPtr, fmt) ;
	vsprintf(buff, fmt, argPtr) ;
	va_end(argPtr) ;

	if(g_LogFilePtr)
	{
		fprintf(g_LogFilePtr, "%s", buff) ;
		fflush(g_LogFilePtr) ;
	}
	else
	{
		// just in case there is some logical error in this app
		CloseLogFile() ;

		g_LogFilePtr = fopen(g_lpszLogFileName, "w") ;
	}
}

void CloseLogFile()
{
	if(g_LogFilePtr)
	{
		fclose(g_LogFilePtr) ;
		g_LogFilePtr = 0 ;
	}
}


BOOL InitVertexBuffer(LPDIRECT3DVERTEXBUFFER8* ppVB, 
					  const IDirect3DDevice8* device8, void* sourcePtr, const UINT cbSize, 
					  const DWORD dwVertexFormat, const DWORD dwUsage, const D3DPOOL d3dpool)
{
	BOOL						res = TRUE ;
	LPDIRECT3DVERTEXBUFFER8		vb = 0 ;
	HRESULT						hr ;
	VOID*						pVertices = 0 ;

	hr = device8->lpVtbl->CreateVertexBuffer(	(IDirect3DDevice8*)device8, cbSize, dwUsage, 
												dwVertexFormat, d3dpool, &vb) ;
	if(FAILED(hr))
    {
        GetDXErrorEx(" FAILED to create vertex buffer. ", hr) ;
		e_xit ;
    }

	hr = vb->lpVtbl->Lock(vb, 0, cbSize, (BYTE**)&pVertices, 0) ;
    if(FAILED(hr))
	{
		GetDXErrorEx(" FAILED to Lock vertex buffer. ", hr) ;
		e_xit ;
	}
    memcpy(pVertices, sourcePtr, cbSize);
    vb->lpVtbl->Unlock(vb) ;

xit :
	if(!res)
	{
		SAFE_RELEASE(vb) ;
	}
	*ppVB = vb ;
	return res ;
}


BOOL InitIndexBuffer(LPDIRECT3DINDEXBUFFER8* ppIB, 
					  const IDirect3DDevice8* device8, void* sourcePtr, const UINT cbSize, 
					  const DWORD dwVertexFormat, const DWORD dwUsage, const D3DPOOL d3dpool)
{
	BOOL						res = TRUE ;
	LPDIRECT3DINDEXBUFFER8		ib = 0 ;
	HRESULT						hr ;
	VOID*						pFaces = 0 ;

	hr = device8->lpVtbl->CreateIndexBuffer(	(IDirect3DDevice8*)device8, cbSize, dwUsage, 
												dwVertexFormat, d3dpool, &ib) ;
	if(FAILED(hr))
    {
        GetDXErrorEx(" FAILED to create vertex buffer. ", hr) ;
		e_xit ;
    }

	hr = ib->lpVtbl->Lock(ib, 0, cbSize, (BYTE**)&pFaces, 0) ;
    if(FAILED(hr))
	{
		GetDXErrorEx(" FAILED to Lock vertex buffer. ", hr) ;
		e_xit ;
	}
    memcpy(pFaces, sourcePtr, cbSize);
    ib->lpVtbl->Unlock(ib) ;

xit :
	if(!res)
	{
		SAFE_RELEASE(ib) ;
	}
	*ppIB = ib ;
	return res ;
}


void __fastcall SetYRotate(D3DMATRIX* mat, const float angle)
{
	float	costh = (float)cos(angle) ;
	float	sinth = (float)sin(angle) ;

	mat->_11 = mat->_33 = costh ;
	mat->_31 = sinth ;
	mat->_13 = -sinth ;
}
